unit FolderTree;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ShlObj, ShellAPI;

const
  WM_DIRCHANGE = WM_USER + 0;
 (* Events of WM_DEVICECHANGE (wParam) *)

  DBT_DEVICEARRIVAL = $8000;            (* system detected a new device *)
  DBT_DEVICEQUERYREMOVE = $8001;        (* wants to remove, may fail *)
  DBT_DEVICEQUERYREMOVEFAILED = $8002;  (* removal aborted *)
  DBT_DEVICEREMOVEPENDING = $8003;      (* about to remove, still avail *)
  DBT_DEVICEREMOVECOMPLETE = $8004;     (* device is gone *)
  DBT_DEVICETYPESPECIFIC = $8005;       (* type specific event *)
  DBT_CONFIGCHANGED = $0018;

 (* type of device in DEV_BROADCAST_HDR *)
  DBT_DEVTYP_OEM = $00000000;           (* OEM- or IHV-defined *)
  DBT_DEVTYP_DEVNODE = $00000001;       (* Devnode number *)
  DBT_DEVTYP_VOLUME = $00000002;        (* Logical volume *)
  DBT_DEVTYP_PORT = $00000003;          (* Port (serial or parallel *)
  DBT_DEVTYP_NET = $00000004;           (* Network resource *)

  (* media types in DBT_DEVTYP_VOLUME *)
  DBTF_MEDIA = $0001;
  DBTF_NET = $0002;                     (* logical volume is network volume *)

type
 PDEV_BROADCAST_HDR = ^TDEV_BROADCAST_HDR;
 TDEV_BROADCAST_HDR = packed record
   dbch_size : DWORD;
   dbch_devicetype : DWORD;
   dbch_reserved : DWORD;
  end;

  PDEV_BROADCAST_VOLUME = ^TDEV_BROADCAST_VOLUME;
  TDEV_BROADCAST_VOLUME = packed record
   dbcv_size : DWORD;
   dbcv_devicetype : DWORD;
   dbcv_reserved : DWORD;
   dbcv_unitmask : DWORD;
   dbcv_flags : WORD;
  end;

  TWMDeviceChange = record
   Msg : Cardinal;
   Event : UINT;
   dwData : Pointer;
   Result : LongInt;
  end;

  TWM_DIRCHANGE = record
    Msg    : word;         { first is the message ID }
    fill   : integer;      { unused }
    Node   : TTreeNode;    { Node whose structure changed }
    result : longint;
  end;

type
  TPathInfo = record
     Path      : string;
     IconIndex : uInt;
     SelectedIndex : uInt;
     Pid           : PItemIdList;
  end;

type
  TDisplayMode = (dmAsis,dmUppercase,dmLowercase);

type
  TFolderTree = class(TCustomTreeView)
  private
     DesktopNode     : TTreeNode;
     MyComputerNode  : TTreeNode;
     SysImageList    : TImageList;
     fDesktopPath    : string;
     fMessageHandle : hWnd;
     fDisplayMode         : TDisplayMode;
     function  fGetDesktop         : TPathInfo;
     function  fGetDesktopFolder   : TPathInfo;
     function  fGetMyComputer      : TPathInfo;
     function  fGetSpecialFolder(nFolder : integer) : TPathInfo;
     function  CreateNode(path : string; parent : TTreeNode) : TTreeNode;
     procedure RetrieveSysImageList;
     function  GetLongFilePath(const FileName : String):string;
     function  AttachFolders(var Node : TTreeNode) : boolean;
     function  GetVol_Ser (drive : string) : dword;
     function  FindPath(path : string) : TTreeNode;
     procedure MessageProc(Var Message : TMessage);
     procedure StartChangeThread(Node : TTreeNode);
     procedure FreeChildren(Node : TTreeNode);
     procedure FreeNode(Node : TTreeNode);
     procedure DriveChanged(Sender: TObject; FirstDriveLetter: Char);
     procedure TFolderTreeExpanding(Sender: TObject; Node: TTreeNode;
         var AllowExpansion: Boolean);
     procedure TFolderTreeClick(Sender: TObject);
     function  GetDirectory : string;
     function GetFirstDriveLetter(unitmask:longint):char;
     procedure WMDeviceChange(var Msg : TWMDeviceChange); dynamic;
    { Private declarations }
  protected
    { Protected declarations }
  public
    property Items;
    property Selected;
    property Directory : string read GetDirectory;
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    function SetDirectory(path : string) : TTreeNode;
    function  GetPath (Node : TTreeNode) : string;
    { Public declarations }
  published
    property OnChange;
    property OnDragOver;
    property OnDragDrop;
    property DisplayMode : TDisplayMode read fDisplayMode write fDisplayMode
                               default dmAsis;
    property ShowButtons;
    property ShowLines;
    property ShowRoot;
    property HideSelection;
    property Align;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property Indent;
    property OnEnter;
    property OnExit;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Visible;
    { Published declarations }
  end;

type
  TThDirChange = class(TThread)
  private
  protected
    procedure Execute; override;
  private
    procedure DoOnChange;
  public
    Node       :  TTreeNode;
    FolderTree :  TFolderTree;
    msghandle  :  hwnd;
  end;

type
  TFolderNode = record
     FN_Path         : string;
     FN_Has_Parent   : boolean;
     FN_Change_Thread : TThDirChange;
     FN_Vol_Ser       : dword;
end;

procedure Register;

implementation

constructor TFolderTree.Create(AOwner : TComponent);
const
   alphabet = '      zyxwvutsrqponmlkjihgfedcba';
var
   drivebits       : longint;
   compbit         : longint;
   i               : byte;
   nodeptr         : ^TFolderNode;
   pathinfo        : TPathInfo;
   displayname     : string;
   drivenode       : TTreeNode;
   foldernode      : ^TFolderNode;
   path            : string;
   AllowExpansion  : boolean;

begin
   inherited Create(AOwner);
   self.parent      := (AOwner as TForm);
   fMessageHandle := AllocateHWnd(MessageProc);
   self.OnExpanding := TFolderTreeExpanding;
   self.OnClick     := TFolderTreeClick;
   self.ReadOnly    := true;

   // Get the system image list.
   RetrieveSysImageList;
   with SysImageList do begin
      masked   := false;
      shareimages := true;
   end;
   self.Images := SysImageList;

   // Establish the node for the Desktop.
   pathinfo := fGetDesktop;
   fDesktopPath := pathinfo.path;
   DesktopNode := CreateNode(pathinfo.path,nil);     // Create a node for the Desktop
   displayname := GetLongFilePath(pathinfo.path);
   with DesktopNode do begin
        imageindex := pathinfo.iconindex;
        selectedindex := pathinfo.selectedindex;

        // Set FN_Path to the path name of the desktop.
        foldernode := data;
        with foldernode^ do
             FN_Path       := pathinfo.path;
   end;

   // Add folders on the desktop to the Desktop node.
   AttachFolders(DesktopNode);

   // Establish the node for My Computer;
   pathinfo := fGetMyComputer;
   MyComputerNode := CreateNode(pathinfo.path,DesktopNode);
   displayname := GetLongFilePath(pathinfo.path);
   with MyComputerNode do begin
        text := displayname;
        imageindex := pathinfo.iconindex;
        selectedindex := pathinfo.selectedindex;
        foldernode := MyComputerNode.data;
        with foldernode^ do
             FN_Has_Parent := false;
   end;

   // Add nodes for each drive in the system. Kill the parent pointer,
   // so the routine that builds complete path names doesn't go higher
   // in the tree.
   drivebits := GetLogicalDrives;
   compbit   := 1;
   for i := 32 downto 7 do begin
       if drivebits and compbit <> 0
          then begin
               path := copy(alphabet,i,1) + ':\';
               drivenode := CreateNode(path,MyComputerNode);
               drivenode.text := GetLongFilePath(path);

               // Indicate the node has no parent (i.e., it's the top of the
               // hierarchy for path construction), and cut off the name at
               // the drive plus a colon, since path construction will add
               // the backslash.
               foldernode := drivenode.data;
               with foldernode^ do begin
                    FN_Path       := copy(alphabet,i,1) + ':';
                    FN_Has_Parent := false;
               end;
       end;
       compbit := compbit shl 1;
   end;

   // Expand the Desktop and My Computer nodes.
   DesktopNode.expand(false);
   MyComputerNode.expand(false);
   Application.ProcessMessages;
end;

function  TFolderTree.fGetDesktop   : TPathInfo;
begin
    result := fGetSpecialFolder(CSIDL_DESKTOP);
end;

function  TFolderTree.fGetMyComputer   : TPathInfo;
begin
    result := fGetSpecialFolder(CSIDL_DRIVES);
end;

function  TFolderTree.fGetDesktopFolder   : TPathInfo;
begin
    result := fGetSpecialFolder(CSIDL_DESKTOPDIRECTORY);
end;

function TFolderTree.fGetSpecialFolder(nFolder : integer) : TPathInfo;
var
  aPidl: PItemIDList;
  TC     : TComponent;
  fLinkDir : string;
  FileInfo  : TSHFileInfo;
begin
  // Get the folder location (as a PItemIDList)
  if SUCCEEDED(SHGetSpecialFolderLocation(self.parent.handle, nFolder, aPidl))
     then begin
       result.Pid := aPidl;

    // Get the actual path of the desktop directory from the PItemIDList
       SetLength(fLinkDir, MAX_PATH); // SHGetPathFromIDList assumes MAX_PATH buffer
       SHGetPathFromIDList(aPidl, PChar(fLinkDir)); // Do it
       SetLength(fLinkDir, StrLen(PChar(fLinkDir)));
       Result.Path := fLinkDir;

       SHGetFileInfo(Pointer(aPidl),SFGAO_SHARE, FileInfo, sizeof(FileInfo),
              SHGFI_PIDL or SHGFI_SYSICONINDEX );
       Result.IconIndex:=FileInfo.iIcon;

       SHGetFileInfo(Pointer(aPidl),SFGAO_SHARE, FileInfo, sizeof(FileInfo),
              SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_OPENICON);
       Result.SelectedIndex:=FileInfo.iIcon;
  end;
end;

// CreateNode creates a Folder_Node and inserts it under
// the "parent" node (if any), using the last of the path
// string as the Name, and setting the new node's text
// property to match.
function TFolderTree.CreateNode(path : string; parent : TTreeNode) : TTreeNode;
var
   filename  : string;
   newfolder : ^TFolderNode;
   dataptr   : ^TFolderNode;
   FileInfo  : TSHFileInfo;
   ImageIndex: Uint;
   SelectedIndex : Uint;
   newnode   : TTreeNode;
   pidl      : PItemIdList;
   displayname : string;
begin
   try
      filename := ExtractFileName(path);
      if filename = ''
         then filename := copy(path,1,1);

      SHGetFileInfo(pchar(path),SFGAO_SHARE, FileInfo, sizeof(FileInfo),
          SHGFI_SYSICONINDEX );
      ImageIndex:=FileInfo.iIcon;

      SHGetFileInfo(pchar(path),SFGAO_SHARE, FileInfo, sizeof(FileInfo),
          SHGFI_SYSICONINDEX or SHGFI_OPENICON);
      SelectedIndex:=FileInfo.iIcon;

      new(newfolder);
      with newfolder^ do begin
          FN_Path          := filename;
          FN_Has_Parent    := (parent <> nil);
          FN_Change_Thread := nil;
          FN_Vol_Ser       := 1;
      end;

      // Add the new node under the parent, if any.
      // If inserting My Computer, insert it as the first child node.
      if parent <> nil
         then if filename = ''
                  then NewNode := self.Items.AddChildObjectFirst
                                                (parent,filename,newfolder)
                  else NewNode := self.Items.AddChildObject
                                                (parent,filename,newfolder)
         else NewNode := self.Items.AddObjectFirst
                       (self.selected,filename,newfolder);
      result := NewNode;
      NewNode.ImageIndex := ImageIndex;
      NewNode.SelectedIndex := SelectedIndex;
      NewNode.HasChildren := true;
      case fDisplayMode of
           dmAsis           : NewNode.text := filename;
           dmLowercase      : NewNode.text := lowercase(filename);
           dmUppercase      : NewNode.text := uppercase(filename);
      end;
   finally
   end;
end;

procedure TFolderTree.RetrieveSysImageList;
Var
  FileInfo : TSHFileInfo;
  DesktopInfo       : TPathInfo;
  DesktopName       : string;
begin
  DesktopInfo := fGetDesktopFolder;
  DesktopName := DesktopInfo.Path;
  SysImageList := TImageList.create(self);
  SysImageList.shareimages := true;
  SysImageList.handle := SHGetFileInfo(PChar(DesktopName), 0, FileInfo, sizeof(FileInfo),
                       SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
end;

Function TFolderTree.GetLongFilePath(const FileName : String) : string;
var
    ShellFolder : IShellFolder ;
    DispName    : array [0..MAX_PATH] of widechar;
    PEaten      : Ulong;
    PIdList     : PitemIdList;
    PAttrib     : Ulong;
    StrReturn   : TStrRet;
begin
     // Obtain a pointer to an IshellFolder interface
     // If FileName is non-null, use ParseDisplayName to
     // obtain the PIdList. Otherwise, it's
     // My Computer--call hGetMyComputer to get the PIdList.
     SHGetDesktopFolder(ShellFolder);
     if FileName <> ''
        then begin
             StringToWideChar(FileName,DispName,MAX_PATH);
             ShellFolder.ParseDisplayName(Application.Handle, nil, DispName,
                             PEaten,PIdList,PAttrib);
        end
        else PidList := fGetMyComputer.Pid;

     ShellFolder.GetDisplayNameOf(PidList,SHGDN_NORMAL,StrReturn);
     Result:=String(StrReturn.cStr);
//     case fDisplayMode of
//           dmLowercase      : result := lowercase(result);
//           dmUppercase      : result := uppercase(result);
//     end;
     ShellFolder.release;
end;


// GetDirectory is the read method for the Directory property.
// It calls GetPath using the currently selected node to build
// the current path.
function TFolderTree.GetDirectory : string;
var
    Node       :  TTreeNode;
    foldernode : ^TFolderNode;
begin
    Node := self.selected;
    if Node = nil
       then result := ''
       else begin
            foldernode := Node.data;
            if (foldernode^.FN_Vol_Ser < 0)        // If the disk isn't loaded,
                   and (foldernode^.FN_Has_Parent) // and not selecting the root,
               then result := ''                   // return a null string.
               else result := GetPath(Node);
       end;
end;

// SetDirectory calls FindPath to get the node closest to the path
// passed to it, then selects the Node in the tree and returns it.
function TFolderTree.SetDirectory(path : string) : TTreeNode;
var
   Node : TTreeNode;
begin
   if path <> ''
       then Node := FindPath(path)
       else Node := nil;
   self.selected := Node;
   result        := Node;
end;


// When the user tries to expand a node with no children, add its
// subfolders as children.
procedure TFolderTree.TFolderTreeExpanding(Sender: TObject; Node: TTreeNode;
  var AllowExpansion: Boolean);
var
  foldernode : ^TFolderNode;
begin
  // If Node already has children, we don't have to find them,
  // and we don't have to start the change thread.
  AllowExpansion := false;
  if Node.count > 0
     then begin
          AllowExpansion := true;
          exit;
          end;

  foldernode := Node.data;

  // If this is the My Computer node, there's nothing to expand.
  if foldernode^.FN_Path = ''
     then begin
          AllowExpansion := true;
          exit;
          end;

  AllowExpansion := AttachFolders(Node);
  Node.HasChildren    := AllowExpansion;
end;

// AttachFolders unconditionally attaches child folders to the
// node passed in Node. If there are any, it also starts a thread
// to watch for changes in the tree structure.
function TFolderTree.AttachFolders(var Node : TTreeNode) : boolean;
var
  foldernode : ^TFolderNode;
  newnode    : TTreeNode;
  newfolder  : ^TFolderNode;
  path       : string;
  ff         : word;
  sr         : TSearchRec;
  retry      : integer;
  errmsg     : string;
  vol_ser    : dword;
  swapped    : boolean;
begin;
   if ComponentState <> []
      then begin
           result := true;
           exit;
           end;
  result := false;
  swapped := false;
  path  := GetPath(Node);    // Get the complete path for this node.
  foldernode := Node.data;

  // If the node being expanded is a root, get the volume serial number
  // and lop off the backslash.
  if length(path) <= 3 then begin
      delete(path,length(path),1);
      vol_ser := GetVol_Ser(path);
      foldernode^.FN_Vol_Ser := vol_ser;
  end;

  vol_ser := foldernode^.FN_Vol_Ser;

  // Add a node for each directory underneath the selected node.
  repeat
      ff := FindFirst (Path + '\*.*', faDirectory, sr);
      if ff = 0
         then retry := IDOK
         else begin
              if ff = 21
                 then begin
                        errmsg := 'Drive ' + copy(Path,1,2) + ' is not ready.';
                        retry  := Application.MessageBox
                                  (pchar(errmsg),
                                  'Error',
                                  MB_ICONERROR + MB_RETRYCANCEL);
                 end

                 // If the directory structure has changed unexpectedly,
                 // (and we weren't notified), the user must have swapped
                 // disks. Climb back to the root, perform DriveChanged,
                 // and select the root.
                 else begin
                        Node := self.selected;
                        foldernode := Node.data;
                        while foldernode^.FN_Has_Parent do begin
                            Node := Node.Parent;
                            foldernode := Node.data;
                        end;
                        DriveChanged(self, path[1]);  // Collapse the root
                        Path := lowercase(foldernode^.FN_Path);
                        foldernode^.FN_Vol_Ser := GetVol_Ser(path);
                        self.selected := Node;
                        retry := IDRETRY;
                        swapped := true;
                 end;
         end;
  until retry <> IDRETRY;

  while ff = 0 do begin
      if (sr.attr and fadirectory) = fadirectory
         then begin
              if (sr.name = '.') or (sr.name = '..')
                         or (lowercase(Path + '\' + sr.name) =
                               lowercase(fDesktopPath))
                 then
                 else begin
                      newnode := CreateNode(Path + '\' + sr.name, Node);
                      newfolder := newnode.data;
                      newfolder^.FN_Vol_Ser := vol_ser;
                      result := true;
                      end;
      end;
      ff := FindNext (sr)
   end;
   FindClose(sr);

   Node.AlphaSort;

   // If result = true (i.e., some children were found), start
   // a thread to watch for changes, unless there already is one.
   if result = true
      then begin
           if foldernode.FN_Change_Thread = nil
              then StartChangeThread(Node);
   end;
end;

// This routine is used only for diskettes. It returns the volume serial
// number of a disk in the drive--if there is no disk there, it returns
// -1. It's used whenever the user clicks on a node, to determine whether
// he has swapped diskettes, necessitating a discard of all nodes for the
// affected drive.
// For volumes other than diskettes, it returns a dummy value of 1,
// as -1 is used to signify a diskette drive that isn't loaded.
function TFolderTree.GetVol_Ser (drive : string) : dword;
var
  RootPathName       : string;
  VolumeSerialNumber : DWORD;
  MaximumComponentLength : DWORD;
  FileSystemFlags : DWORD;
  SerialNumber : string;
  driveletter  : char;
  ErrorMode    : word;
  path         : string;
begin
  result := -1;

  // If there is no path, it's My Computer--exit now.
  if length(drive) = 0 then exit;

  // If the drive isn't removeable, we don't have to do this.
  path := drive[1] + ':\';
  path := lowercase(path);
  if GetDriveType(pchar(path)) <> DRIVE_REMOVABLE
       then begin
               result := 1;
               exit;
            end;

  // If there's no disk in the drive, bail out with a result = -1.
  { make it upper case }
  path := uppercase(path);
  driveletter := path[1];
  { turn off critical errors }
  ErrorMode := SetErrorMode(SEM_FailCriticalErrors);
  try
    { drive 1 = a, 2 = b, 3 = c, etc. }
    if DiskSize(Ord(Driveletter) - $40) = -1
      then exit;
  finally
    { restore old error mode }
    SetErrorMode(ErrorMode);
  end;

  // Get the volume serial number of the disk.
  RootPathName := uppercase(copy(drive,1,1)) + ':\';
  GetVolumeInformation(PChar(RootPathName),
                       nil,
                       0,
                       @VolumeSerialNumber,
                       MaximumComponentLength,
                       FileSystemFlags,
                       nil,
                       0);

  result := integer(VolumeSerialNumber);
end;

// GetPath works its way up the tree to find construct the path leading
// to Node.
function TFolderTree.GetPath (Node : TTreeNode) : string;
var
  path       : string;
  foldernode : ^TFolderNode;
  worknode   : TTreeNode;
  pnode      : TTreeNode;
begin
  path := '';
  if not assigned(Node)
     then exit;

  worknode := Node;
  foldernode := worknode.data;
  with foldernode^ do begin
      path := FN_Path;
      if FN_Has_Parent then
         repeat
            path := '\' + path;
            pnode := worknode.parent;
            worknode := pnode;
            foldernode := worknode.data;
            path       := foldernode^.FN_Path + path;
         until not foldernode^.FN_Has_Parent;
  end;

  if length(path) > 0                   // If path is null, it's My Computer.
     then if length(path) < 3           // If the result is less than 3
             then path := path + '\';   // bytes, its the root--add a "\".
  result := path;
end;

// FindPath tries to find the path passed in "path" in the tree,
// and returns the node closest to it.
function TFolderTree.FindPath(path : string) : TTreeNode;
var
   folder       : TStringList;
   workpath     : string;
   tail         : string;
   Node         : TTreeNode;
   LastNode     : TTreeNode;
   FolderNode   : ^TFolderNode;
   depth        : integer;
begin
   folder := TStringList.create;
   result := nil;

   // Break "path" up into individual levels, which are stored in
   // the folder list in LIFO order. If the path starts with the
   // desktop folder, we'll pile the whole desktop folder path
   // onto the stack at once.
   workpath := lowercase(path);
   if length(workpath) < 3
      then begin
           self.selected := nil;
           exit;
           end;
   if copy(workpath,length(workpath),1) = '\'    // Lop off any trailing
      then delete(workpath,length(workpath),1);  // backslash.
   while length(workpath) > 0 do begin
      if workpath = lowercase(fDesktopPath)
         then tail   := workpath
         else begin
              tail   := ExtractFileName(workpath);
              if tail = '' then tail := workpath;
              end;
      folder.insert(0,tail);
      workpath := copy(workpath,1,(length(workpath) - length(tail) - 1));
   end;

   // Locate each successive element of the path, one level
   // at a time, except: the first element must be found
   // at DesktopNode or in a child of MyComputerNode.
   workpath := folder[0];            // Set workpath to the first path node.
   Node := DesktopNode;
   FolderNode := Node.data;
   if lowercase(FolderNode^.FN_Path) = workpath
                                     // If the first path node is the desktop
      then                           // then continue processing...
      else begin
           Node := MyComputerNode.GetFirstChild;
           while Node <> nil do begin
               FolderNode := Node.data;                      // Search children of
               if lowercase(FolderNode^.FN_Path) = workpath  // MyComputerNode for a match.
                  then break
                  else Node := MyComputerNode.GetNextChild(Node);
           end;
      end;

   // If no match is found, give up.
   if Node = nil
      then self.selected := nil

   // Starting with the children of the current Node, look for the
   // next level. Repeat until there is no match, or the list of
   // nodes is exhausted.
      else begin
           LastNode := Node;
           depth    := 1;
           while depth < Folder.count do begin
              if Node.count <= 0
                 then AttachFolders(LastNode);     // Ensure children are found.
              Node := LastNode.GetFirstChild;
              workpath := Folder[depth];
              while Node <> nil do begin
                 FolderNode := Node.data;          // Search children of
                 if lowercase(FolderNode^.FN_Path)
                                       = workpath  // MyComputerNode for a match.
                    then break
                    else Node := MyComputerNode.GetNextChild(Node);
              end;

              // If a matching node was found, set LastNode to that node,
              // then try the next level down, if any.
              if Node <> nil
                 then begin
                      LastNode := Node;
                      inc(depth);
                 end
                 else break;
           end;

           // Return the last matching nod.
           result := LastNode;
      end;

   folder.free;
end;

// MessageProc is the message handling routine.
procedure TFolderTree.MessageProc(var Message: TMessage);
var
  Node           : TTreeNode;
  NodeExpanded   : boolean;
  AllowExpansion : boolean;
  handled        : boolean;
  cursorpos      : TPoint;
begin
  handled := false;

  // Handle file system changes here.
  if (Message.Msg = WM_DIRCHANGE) then begin
     Node := TWM_DIRCHANGE(Message).Node;
     NodeExpanded := Node.expanded;

     // If the node is expanded, collapse it.
     if NodeExpanded
        then Node.collapse(true);
     Application.ProcessMessages;

     // Free all the child nodes of the current node.
     FreeChildren(Node);

     // If the node was expanded before, expand it now. Otherwise,
     // just call the OnExpanding routine to attach child nodes
     // without showing.
     TFolderTreeExpanding(self, Node, AllowExpansion);
     if NodeExpanded
        then Node.expand(false);
     handled := true;
  end;

  if (Message.Msg = WM_DEVICECHANGE) then begin
    try
        WMDeviceChange(TWMDeviceChange(Message));
    except
        Application.HandleException(Self);
    end;
    handled := true;
  end;

  if (Message.Msg = WM_PAINT) then
     if csDesigning in ComponentState
        then begin
               GetCursorPos(cursorpos);
               cursorpos.x := cursorpos.x - parent.left;
               cursorpos.y := cursorpos.y - parent.top;
               top :=  cursorpos.y;
               left := cursorpos.x;
               handled := true;
     end;

  if not handled
     then Message.Result :=
          DefWindowProc(fMessageHandle, Message.Msg,
                     Message.wParam, Message.lParam);
end;

procedure TFolderTree.WMDeviceChange(var Msg : TWMDeviceChange);
var lpdb : PDEV_BROADCAST_HDR;
    lpdbv : PDEV_BROADCAST_VOLUME;
begin
 (* received a wm_devicechange message *)
  lpdb := PDEV_BROADCAST_HDR(Msg.dwData);
  (* look at the event send together with the wm_devicechange message *)
   case Msg.Event of
    DBT_DEVICEARRIVAL : begin
      if lpdb^.dbch_devicetype = DBT_DEVTYP_VOLUME then begin
       lpdbv := PDEV_BROADCAST_VOLUME(Msg.dwData);
       if (lpdbv^.dbcv_flags and DBTF_MEDIA) = 1 then
         DriveChanged(Self, GetFirstDriveLetter(lpdbv^.dbcv_unitmask));
      end;
     end;
    DBT_DEVICEREMOVECOMPLETE : begin
      if lpdb^.dbch_devicetype = DBT_DEVTYP_VOLUME then begin
       lpdbv := PDEV_BROADCAST_VOLUME(Msg.dwData);
       if (lpdbv^.dbcv_flags and DBTF_MEDIA) = 1 then
         DriveChanged(Self, GetFirstDriveLetter(lpdbv^.dbcv_unitmask));
      end;
     end;
   end;
 end;

function TFolderTree.GetFirstDriveLetter(unitmask : longint):char;
var DriveLetter : shortint;
begin
 DriveLetter := Ord('A');
 while (unitmask and 1)=0  do begin
  unitmask := unitmask shr 1;
  inc(DriveLetter);
 end;
 Result := Char(DriveLetter);
end;

// This procedure starts a thread to notify us if the
// directory structure changes.
procedure TFolderTree.StartChangeThread(Node : TTreeNode);
var
    foldernode : ^TFolderNode;
    thread     : TThDirChange;
begin
   if ComponentState <> []
      then exit;
    foldernode := Node.data;
    if foldernode^.FN_Change_Thread = nil
       then begin
            thread := TThDirChange.create(true);
            thread.msghandle := fMessageHandle;
            thread.Node  := Node;
            thread.FolderTree := self;
            thread.FreeOnTerminate:=True;
            foldernode^.FN_Change_Thread := thread;
            thread.Resume;
    end;
end;

destructor TFolderTree.Destroy;
begin
    DeAllocateHWnd(fMessageHandle);
    SysImageList.free;
    while self.items.count > 0 do
       FreeNode(self.items[0]);
    inherited;
end;

// FreeChildren frees the descendants of the Node passed to it.
procedure TFolderTree.FreeChildren(Node : TTreeNode);
begin
  while Node.count > 0 do
     FreeNode(Node.item[0]);
end;     

procedure TFolderTree.FreeNode(Node : TTreeNode);
var
    foldernode : ^TFolderNode;
    thread     : TThDirChange;
begin
     while Node.count > 0 do
         FreeNode(Node.item[0]);

     foldernode := Node.data;
     thread := foldernode^.FN_Change_Thread;
     if thread <> nil
          then thread.terminate;
     dispose(foldernode);
     Node.delete;
end;

procedure TFolderTree.DriveChanged(Sender: TObject; FirstDriveLetter: Char);
var
    Node       : TTreeNode;
    path       : string;
    foldernode : ^TFolderNode;
begin
    // Find the node corresponding to the drive that has changed.
    Node := MyComputerNode.item[0];
    while Node <> nil do begin
        foldernode := Node.data;
        if lowercase(foldernode^.FN_Path[1]) = lowercase(FirstDriveLetter)
            then break;
        Node := Node.GetNextSibling;
    end;
    if Node <> nil then begin
        Node.collapse(true);                  // Collapse the node.
        path := FirstDriveLetter + ':\';      // Set "path"
        FreeChildren(Node);                   // Free its children
        Node.HasChildren := true;             // Attach the plus sign button
        Node.text := GetLongFilePath(path);   // Set node's text to long path
    end;

end;

// When the user clicks on a node, and it points to a removeable
// drive, determine whether the disk in the drive has changed.
// If so, perform DriveChanged.
procedure TFolderTree.TFolderTreeClick(Sender: TObject);
var
     Node       : TTreeNode;
     foldernode : ^TFolderNode;
     vol_ser    : dword;
     path       : string;
begin
     if not assigned(self.selected)
         then exit;
     Node := self.selected;
     foldernode := Node.data;

     // If My Computer was selected, exit.
     if length(foldernode^.FN_Path) = 0 then exit;

     // If the node is the root directory of a drive, find its
     // vol_ser number. Otherwise, it's already inherited from
     // that root.
     with foldernode^ do begin
          if FN_Has_Parent
             then vol_ser  := FN_Vol_Ser
             else begin
                  path := lowercase(copy(FN_Path,1,1));
                  vol_ser := GetVol_Ser(path);
             end;
     end;

     // If there was a disk in the drive before (foldernode^.FN_Vol_Ser <> -1),
     // and the number has changed, call DriveChanged.
     if (vol_ser <> foldernode^.FN_Vol_Ser)
                  and (foldernode^.FN_Vol_Ser <> -1)
         then begin
              DriveChanged(self, path[1]);
              foldernode^.FN_Vol_Ser := vol_ser;
              end;
end;

//==============================================================
procedure TThDirChange.Execute;
var
   fr:TWOHandleArray;
   i:integer;
   s:string;
   Directory: string;
begin
     // Find the path that ends in the selected node.
     Directory := FolderTree.GetPath(Node);

     if Copy(Directory,Length(Directory),1) <> '\'
          then Directory:=Directory+'\';
     // If Directory is just a root, add 1 to its length
     // to compensate for the character subtracted from
     // the length when 's' is built.
     if length(Directory) <= 3
          then Directory:=Directory+'\';

     s:=Copy(Directory,1,Length(Directory) - 1)+#0;

     while not Terminated do begin
           fr[0]:=FindFirstChangeNotification
                       (Pchar(s),false,FILE_NOTIFY_CHANGE_DIR_NAME);
           if FindNextChangeNotification(fr[0])
              then begin
                   i:=WAIT_TIMEOUT;
                   while (i=WAIT_TIMEOUT) and (not Terminated) do
                         i:=WaitForMultipleObjects(1,@fr,false,20);
                   if Not Terminated
                       then Synchronize(DoOnChange);
           end;
           FindCloseChangeNotification(fr[0]);
     end;

     DoTerminate;
end;

// On change, send a message to the changed node identifying itself.
procedure TThDirChange.DoOnChange;
begin
   // Find the path that ends in the selected node.
   SendMessage(msghandle,WM_DIRCHANGE,0,longint(Node));
end;

procedure Register;
begin
  RegisterComponents('Custom', [TFolderTree]);
end;

end.
